"""Top-level package for pyMeticulous."""

from .api import Api, APIError
from .profile import Profile

__all__ = ("Api", "APIError", "Profile")

__author__ = """Meticulous"""
__email__ = "support@meticuloushome.com"
